;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((sentence-end-double-space . t)
         (fill-column . 70)))
 (log-edit-mode . ((log-edit-font-lock-gnu-style . t)
                   (log-edit-setup-add-author . t)))
 (change-log-mode . ((add-log-time-zone-rule . t)
                     (fill-column . 74)))
 (diff-mode . ((mode . whitespace)))
 (emacs-lisp-mode . ((indent-tabs-mode . nil)
                     (electric-quote-comment . nil)
                     (electric-quote-string . nil))))
