;; Generated package description from eglot.el  -*- no-byte-compile: t -*-
(define-package "eglot" "1.8.0.20221008.103138" "Client for Language Server Protocol (LSP) servers" '((emacs "26.1") (jsonrpc "1.0.14") (flymake "1.2.1") (project "0.3.0") (xref "1.0.1") (eldoc "1.11.0") (seq "2.23")) :commit "c460869ddeeec9556ce7220318dc13d2b6e2b021" :authors '(("João Távora" . "joaotavora@gmail.com")) :maintainer '("João Távora" . "joaotavora@gmail.com") :keywords '("convenience" "languages") :url "https://github.com/joaotavora/eglot")
